"""This is part of the Python's module to manage the DP-EH600 thermal printer.
Source: https://github.com/BoboTiG/thermalprinter.
"""

from arabic_reshaper import reshape
from bidi import algorithm

#: :meta hide-value:
IRAN_SYSTEM_MAP = {
    # Numerals
    0x06F00: 0x80,  # EXTENDED ARABIC-INDIC DIGIT ZERO
    0x06F10: 0x81,  # EXTENDED ARABIC-INDIC DIGIT ONE
    0x06F20: 0x82,  # EXTENDED ARABIC-INDIC DIGIT TWO
    0x06F30: 0x83,  # EXTENDED ARABIC-INDIC DIGIT THREE
    0x06F40: 0x84,  # EXTENDED ARABIC-INDIC DIGIT FOUR
    0x06F50: 0x85,  # EXTENDED ARABIC-INDIC DIGIT FIVE
    0x06F60: 0x86,  # EXTENDED ARABIC-INDIC DIGIT SIX
    0x06F70: 0x87,  # EXTENDED ARABIC-INDIC DIGIT SEVEN
    0x06F80: 0x88,  # EXTENDED ARABIC-INDIC DIGIT EIGHT
    0x06F90: 0x89,  # EXTENDED ARABIC-INDIC DIGIT NINE
    # Punctuations
    0x060C: 0x8A,  # ARABIC COMMA
    0x0640: 0x8B,  # ARABIC TATWEEL
    0x061F: 0x8C,  # ARABIC QUESTION MARK
    # Letters
    0xFE81: 0x8D,  # ARABIC LETTER ALEF WITH MADDA ABOVE ISOLATED FORM
    0xFE8B: 0x8E,  # ARABIC LETTER YEH WITH HAMZA ABOVE INITIAL FORM
    0xFE80: 0x8F,  # ARABIC LETTER HAMZA ISOLATED FORM
    0xFE8D: 0x90,  # ARABIC LETTER ALEF ISOLATED FORM
    0xFE8E: 0x91,  # ARABIC LETTER ALEF FINAL FORM
    0xFE8F: 0x92,  # ARABIC LETTER BEH ISOLATED FORM
    0xFE90: 0x92,
    0xFE91: 0x93,  # ARABIC LETTER BEH INITIAL FORM
    0xFE92: 0x93,
    0xFB56: 0x94,  # ARABIC LETTER PEH ISOLATED FORM
    0xFB57: 0x94,
    0xFB58: 0x95,  # ARABIC LETTER PEH INITIAL FORM
    0xFB59: 0x95,
    0xFE95: 0x96,  # ARABIC LETTER TEH ISOLATED FORM
    0xFE96: 0x96,
    0xFE97: 0x97,  # ARABIC LETTER TEH INITIAL FORM
    0xFE98: 0x97,
    0xFE99: 0x98,  # ARABIC LETTER THEH ISOLATED FORM
    0xFE9A: 0x98,
    0xFE9B: 0x99,  # ARABIC LETTER THEH INITIAL FORM
    0xFE9C: 0x99,
    0xFE9D: 0x9A,  # ARABIC LETTER JEEM ISOLATED FORM
    0xFE9E: 0x9A,
    0xFE9F: 0x9B,  # ARABIC LETTER JEEM INITIAL FORM
    0xFEA0: 0x9A,
    0xFB7A: 0x9C,  # ARABIC LETTER TCHEH ISOLATED FORM
    0xFB7B: 0x9C,
    0xFB7C: 0x9D,  # ARABIC LETTER TCHEH INITIAL FORM
    0xFB7D: 0x9D,
    0xFEA1: 0x9E,  # ARABIC LETTER HAH ISOLATED FORM
    0xFEA2: 0x9E,
    0xFEA3: 0x9F,  # ARABIC LETTER HAH INITIAL FORM
    0xFEA4: 0x9F,
    0xFEA5: 0xA0,  # ARABIC LETTER KHAH ISOLATED FORM
    0xFEA6: 0xA0,
    0xFEA7: 0xA1,  # ARABIC LETTER KHAH INITIAL FORM
    0xFEA8: 0xA1,
    0xFEA9: 0xA2,  # ARABIC LETTER DAL ISOLATED FORM
    0xFEAA: 0xA2,
    0xFEAB: 0xA3,  # ARABIC LETTER THAL ISOLATED FORM
    0xFEAC: 0xA3,
    0xFEAD: 0xA4,  # ARABIC LETTER REH ISOLATED FORM
    0xFEAE: 0xA4,
    0xFEAF: 0xA5,  # ARABIC LETTER ZAIN ISOLATED FORM
    0xFEB0: 0xA5,
    0xFB8A: 0xA6,  # ARABIC LETTER JEH ISOLATED FORM
    0xFB8B: 0xA6,
    0xFEB1: 0xA7,  # ARABIC LETTER SEEN ISOLATED FORM
    0xFEB2: 0xA7,
    0xFEB3: 0xA8,  # ARABIC LETTER SEEN INITIAL FORM
    0xFEB4: 0xA8,
    0xFEB5: 0xA9,  # ARABIC LETTER SHEEN ISOLATED FORM
    0xFEB6: 0xA9,
    0xFEB7: 0xAA,  # ARABIC LETTER SHEEN INITIAL FORM
    0xFEB8: 0xAA,
    0xFEB9: 0xAB,  # ARABIC LETTER SAD ISOLATED FORM
    0xFEBA: 0xAB,
    0xFEBB: 0xAC,  # ARABIC LETTER SAD INITIAL FORM
    0xFEBC: 0xAC,
    0xFEBD: 0xAD,  # ARABIC LETTER DAD ISOLATED FORM
    0xFEBE: 0xAD,
    0xFEBF: 0xAE,  # ARABIC LETTER DAD INITIAL FORM
    0xFEC0: 0xAE,
    0xFEC1: 0xAF,  # ARABIC LETTER TAH ISOLATED FORM
    0xFEC2: 0xAF,
    0xFEC3: 0xAF,
    0xFEC4: 0xAF,
    # Shadows
    0x2591: 0xB0,  # LIGHT SHADE
    0x2592: 0xB1,  # MEDIUM SHADE
    0x2593: 0xB2,  # DARK SHADE
    # Box Drawings
    0x2502: 0xB3,  # BOX DRAWINGS LIGHT VERTICAL
    0x2524: 0xB4,  # BOX DRAWINGS LIGHT VERTICAL AND LEFT
    0x2561: 0xB5,  # BOX DRAWINGS VERTICAL SINGLE AND LEFT DOUBLE
    0x2562: 0xB6,  # BOX DRAWINGS VERTICAL DOUBLE AND LEFT SINGLE
    0x2556: 0xB7,  # BOX DRAWINGS DOWN DOUBLE AND LEFT SINGLE
    0x2555: 0xB8,  # BOX DRAWINGS DOWN SINGLE AND LEFT DOUBLE
    0x2563: 0xB9,  # BOX DRAWINGS DOUBLE VERTICAL AND LEFT
    0x2551: 0xBA,  # BOX DRAWINGS DOUBLE VERTICAL
    0x2557: 0xBB,  # BOX DRAWINGS DOUBLE DOWN AND LEFT
    0x255D: 0xBC,  # BOX DRAWINGS DOUBLE UP AND LEFT
    0x255C: 0xBD,  # BOX DRAWINGS UP DOUBLE AND LEFT SINGLE
    0x255B: 0xBE,  # BOX DRAWINGS UP SINGLE AND LEFT DOUBLE
    0x2510: 0xBF,  # BOX DRAWINGS LIGHT DOWN AND LEFT
    0x2514: 0xC0,  # BOX DRAWINGS LIGHT UP AND RIGHT
    0x2534: 0xC1,  # BOX DRAWINGS LIGHT UP AND HORIZONTAL
    0x252C: 0xC2,  # BOX DRAWINGS LIGHT DOWN AND HORIZONTAL
    0x251C: 0xC3,  # BOX DRAWINGS LIGHT VERTICAL AND RIGHT
    0x2500: 0xC4,  # BOX DRAWINGS LIGHT HORIZONTAL
    0x253C: 0xC5,  # BOX DRAWINGS LIGHT VERTICAL AND HORIZONTAL
    0x255E: 0xC6,  # BOX DRAWINGS VERTICAL SINGLE AND RIGHT DOUBLE
    0x255F: 0xC7,  # BOX DRAWINGS VERTICAL DOUBLE AND RIGHT SINGLE
    0x255A: 0xC8,  # BOX DRAWINGS DOUBLE UP AND RIGHT
    0x2554: 0xC9,  # BOX DRAWINGS DOUBLE DOWN AND RIGHT
    0x2569: 0xCA,  # BOX DRAWINGS DOUBLE UP AND HORIZONTAL
    0x2566: 0xCB,  # BOX DRAWINGS DOUBLE DOWN AND HORIZONTAL
    0x2560: 0xCC,  # BOX DRAWINGS DOUBLE VERTICAL AND RIGHT
    0x2550: 0xCD,  # BOX DRAWINGS DOUBLE HORIZONTAL
    0x256C: 0xCE,  # BOX DRAWINGS DOUBLE VERTICAL AND HORIZONTAL
    0x2567: 0xCF,  # BOX DRAWINGS UP SINGLE AND HORIZONTAL DOUBLE
    0x2568: 0xD0,  # BOX DRAWINGS UP DOUBLE AND HORIZONTAL SINGLE
    0x2564: 0xD1,  # BOX DRAWINGS DOWN SINGLE AND HORIZONTAL DOUBLE
    0x2565: 0xD2,  # BOX DRAWINGS DOWN DOUBLE AND HORIZONTAL SINGLE
    0x2559: 0xD3,  # BOX DRAWINGS UP DOUBLE AND RIGHT SINGLE
    0x2558: 0xD4,  # BOX DRAWINGS UP SINGLE AND RIGHT DOUBLE
    0x2552: 0xD5,  # BOX DRAWINGS DOWN SINGLE AND RIGHT DOUBLE
    0x2553: 0xD6,  # BOX DRAWINGS DOWN DOUBLE AND RIGHT SINGLE
    0x256B: 0xD7,  # BOX DRAWINGS VERTICAL DOUBLE AND HORIZONTAL SINGLE
    0x256A: 0xD8,  # BOX DRAWINGS VERTICAL SINGLE AND HORIZONTAL DOUBLE
    0x2518: 0xD9,  # BOX DRAWINGS LIGHT UP AND LEFT
    0x250C: 0xDA,  # BOX DRAWINGS LIGHT DOWN AND RIGHT
    # Shadows
    0x2588: 0xDB,  # FULL BLOCK
    0x2584: 0xDC,  # LOWER HALF BLOCK
    0x258C: 0xDD,  # LEFT HALF BLOCK
    0x2590: 0xDE,  # RIGHT HALF BLOCK
    0x2580: 0xDF,  # UPPER HALF BLOCK
    # Letters
    0xFEC5: 0xE0,  # ARABIC LETTER ZAH ISOLATED FORM
    0xFEC6: 0xE0,
    0xFEC7: 0xE0,
    0xFEC8: 0xE0,
    0xFEC9: 0xE1,  # ARABIC LETTER AIN ISOLATED FORM
    0xFECA: 0xE2,  # ARABIC LETTER AIN FINAL FORM
    0xFECC: 0xE3,  # ARABIC LETTER AIN MEDIAL FORM
    0xFECB: 0xE4,  # ARABIC LETTER AIN INITIAL FORM
    0xFECD: 0xE5,  # ARABIC LETTER GHAIN ISOLATED FORM
    0xFECE: 0xE6,  # ARABIC LETTER GHAIN FINAL FORM
    0xFED0: 0xE7,  # ARABIC LETTER GHAIN MEDIAL FORM
    0xFECF: 0xE8,  # ARABIC LETTER GHAIN INITIAL FORM
    0xFED1: 0xE9,  # ARABIC LETTER FEH ISOLATED FORM
    0xFED2: 0xE9,
    0xFED3: 0xEA,  # ARABIC LETTER FEH INITIAL FORM
    0xFED4: 0xEA,
    0xFED5: 0xEB,  # ARABIC LETTER QAF ISOLATED FORM
    0xFED6: 0xEB,
    0xFED7: 0xEC,  # ARABIC LETTER QAF INITIAL FORM
    0xFED8: 0xEC,
    0xFB8E: 0xED,  # ARABIC LETTER KEHEH ISOLATED FORM
    0xFB8F: 0xED,
    0xFB90: 0xEE,  # ARABIC LETTER KEHEH INITIAL FORM
    0xFB91: 0xEE,
    0xFB92: 0xEF,  # ARABIC LETTER GAF ISOLATED FORM
    0xFB93: 0xEF,
    0xFB94: 0xF0,  # ARABIC LETTER GAF INITIAL FORM
    0xFB95: 0xF0,
    0xFEDD: 0xF1,  # ARABIC LETTER LAM ISOLATED FORM
    0xFEDE: 0xF1,
    0xFEFB: 0xF2,  # ARABIC LIGATURE LAM WITH ALEF ISOLATED FORM
    0xFEFC: 0xF2,
    0xFEDF: 0xF3,  # ARABIC LETTER LAM INITIAL FORM
    0xFEE0: 0xF3,
    0xFEE1: 0xF4,  # ARABIC LETTER MEEM ISOLATED FORM
    0xFEE2: 0xF4,
    0xFEE3: 0xF5,  # ARABIC LETTER MEEM INITIAL FORM
    0xFEE4: 0xF5,
    0xFEE5: 0xF6,  # ARABIC LETTER NOON ISOLATED FORM
    0xFEE6: 0xF6,
    0xFEE7: 0xF7,  # ARABIC LETTER NOON INITIAL FORM
    0xFEE8: 0xF7,
    0xFEED: 0xF8,  # ARABIC LETTER WAW ISOLATED FORM
    0xFEEE: 0xF8,
    0xFEE9: 0xF9,  # ARABIC LETTER HEH ISOLATED FORM
    0xFEEA: 0xF9,
    0xFEEC: 0xFA,  # ARABIC LETTER HEH MEDIAL FORM
    0xFEEB: 0xFB,  # ARABIC LETTER HEH INITIAL FORM
    0xFBFD: 0xFC,  # ARABIC LETTER FARSI YEH FINAL FORM
    0xFBFC: 0xFD,  # ARABIC LETTER FARSI YEH ISOLATED FORM
    0xFBFE: 0xFE,  # ARABIC LETTER FARSI YEH INITIAL FORM
    0xFBFF: 0xFE,
    0x00A0: 0xFF,  # NO-BREAK SPACE
}

__all__ = (
    "IRAN_SYSTEM_MAP",
    "algorithm",
    "reshape",
)
